package com.yourkit.probes.builtin;

import com.yourkit.probes.*;

import java.awt.event.InputEvent;
import java.awt.event.InvocationEvent;
import java.lang.reflect.Field;

/**
 * Detect long running AWT/Swing events
 */
@SuppressWarnings({"UnusedDeclaration"})
public final class AWTEvents {

  private static final class EventTable extends Table {
    private final ClassNameColumn myEventClass = new ClassNameColumn("Event Class");
    private final ClassNameColumn myRunnableClass = new ClassNameColumn("Runnable Class");

    public EventTable() {
      super(
        "Long AWT Events",
        // It makes no sense to record stack trace

        Table.LASTING_EVENTS |
        Table.RECORD_THREAD_CPU_TIME |
        Table.RECORD_THREAD_ON_ROW_CREATION
      );
    }
  }
  private static final EventTable T_EVENT = new EventTable();
  static {
    // Record only long events
    T_EVENT.setMinimumRecordedLastingEventTime(300);

    // If event lasts long enough, start collecting stack traces more frequently,
    // to learn what is happening.
    // Note, that the threshold of setFrequentSamplingThreshold() is slightly smaller
    // than the threshold passed to setMinimumRecordedLastingEventTime(),
    // to ensure that every retained event has associated stack traces, thus enabling its analysis.
    T_EVENT.setFrequentSamplingThreshold(250);
  }

  @MethodPattern("java.awt.EventQueue:dispatchEvent(java.awt.AWTEvent)")
  public static final class EventQueue_dispatchEvent_Probe {
    public static int onEnter(@Param(1) final Object event) {
      if (event instanceof InputEvent) {
        // these events are of no interest
        return Table.NO_ROW;
      }

      return T_EVENT.createRow();
    }

    public static void onReturn(
      @OnEnterResult final int rowIndex,
      @Param(1) final Object event
    ) {
      processExitImpl(rowIndex, event);
    }

    public static void onUncaughtException(
      @OnEnterResult final int rowIndex,
      @Param(1) final Object event
    ) {
      processExitImpl(rowIndex, event);
    }

    /**
     * Close event whenever it normally finished or terminated by an exception
     */
    private static void processExitImpl(final int rowIndex, final Object event) {
      if (rowIndex < 0) {
        return;
      }

      if (!T_EVENT.closeRow(rowIndex)) {
        // optimization: short event not stored in table - do not set its values
        return;
      }

      T_EVENT.myEventClass.setValue(rowIndex, event.getClass());

      if (event instanceof InvocationEvent) {
        try {
          final Field runnable = InvocationEvent.class.getDeclaredField("runnable");
          runnable.setAccessible(true);
          T_EVENT.myRunnableClass.setValue(rowIndex, runnable.get(event).getClass());
        }
        catch (final Throwable ignored) {
        }
      }
    }
  }
}
