package com.yourkit.probes.builtin;

import com.yourkit.probes.*;

import javax.sql.DataSource;
import javax.sql.PooledConnection;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.Statement;
import java.util.WeakHashMap;

@SuppressWarnings("UnusedDeclaration")
public final class Databases {

  private static final Object ourLock = new Object();

  private static final ObjectRowIndexMap<Connection> ourConnection2RowIndex = new ObjectRowIndexMap<Connection>();
  private static final ObjectRowIndexMap<Connection> ourConnection2TransactionID = new ObjectRowIndexMap<Connection>();
  private static final ObjectRowIndexMap<Statement> ourStatement2RowIndex = new ObjectRowIndexMap<Statement>();
  private static final WeakHashMap<DataSource, String> ourDataSource2Context = new WeakHashMap<DataSource, String>();

  private static final class ConnectionTable extends Table {
    private final StringColumn myDatabase = new StringColumn("Database");

    public ConnectionTable() {
      super("Connections", Table.MASK_FOR_GENERIC_LASTING_EVENTS);
    }
  }
  private static final ConnectionTable T_CONNECTION = new ConnectionTable();

  private static final class TransactionTable extends Table {
    public TransactionTable() {
      super(T_CONNECTION, "Transactions", Table.MASK_FOR_GENERIC_LASTING_EVENTS);
    }
  }
  private static final TransactionTable T_TRANSACTION = new TransactionTable();


  private static final class StatementTable extends Table {
    public StatementTable() {
      super(T_CONNECTION, "Statements", Table.MASK_FOR_GENERIC_LASTING_EVENTS);
    }
  }
  private static final StatementTable T_STATEMENT = new StatementTable();

  private static final class QueryTable extends Table {
    private final StringColumn mySQL = new StringColumn("SQL");

    public QueryTable() {
      super(T_STATEMENT, "Queries", Table.MASK_FOR_SINGLE_METHOD_CALL_LASTING_EVENTS);
    }
  }
  private static final QueryTable T_QUERY = new QueryTable();

  private static final class PreparedStatementTable extends Table {
    private final StringColumn mySQL = new StringColumn("SQL");

    public PreparedStatementTable() {
      super(T_CONNECTION, "Prepared Statements", Table.MASK_FOR_GENERIC_LASTING_EVENTS);
    }
  }
  private static final PreparedStatementTable T_PREPARED_STATEMENT = new PreparedStatementTable();

  private static final class PreparedStatementQueryTable extends Table {
    public PreparedStatementQueryTable() {
      super(T_PREPARED_STATEMENT, "Queries", Table.MASK_FOR_SINGLE_METHOD_CALL_LASTING_EVENTS);
    }
  }
  private static final PreparedStatementQueryTable T_PREPARED_STATEMENT_QUERY = new PreparedStatementQueryTable();


  private static final String BATCH_OPERATION_ADD = "Add";
  private static final String BATCH_OPERATION_EXECUTE = "Execute";
  private static final String BATCH_OPERATION_CLEAR = "Clear";

  private static final class PreparedStatementBatchTable extends Table {
    private final StringColumn myOperation = new StringColumn("Operation");

    public PreparedStatementBatchTable() {
      super(T_PREPARED_STATEMENT, "Batches", Table.MASK_FOR_SINGLE_METHOD_CALL_LASTING_EVENTS);
    }
  }
  private static final PreparedStatementBatchTable T_PREPARED_STATEMENT_BATCH = new PreparedStatementBatchTable();

  private static final class BatchTable extends Table {
    private final StringColumn myOperation = new StringColumn("Operation");
    private final StringColumn mySQL = new StringColumn("SQL");

    public BatchTable() {
      super(T_STATEMENT, "Batches", Table.MASK_FOR_SINGLE_METHOD_CALL_LASTING_EVENTS);
    }
  }
  private static final BatchTable T_BATCH = new BatchTable();


  @MethodPattern("java.sql.DriverManager:getConnection(String, java.util.Properties, java.lang.ClassLoader)")
  public static final class DriverManager_getConnection_Probe {
    public static void onReturn(
      @ReturnValue final Connection connection,
      @Param(1) final String database
    ) {
      final int rowIndex = T_CONNECTION.createRow();
      T_CONNECTION.myDatabase.setValue(rowIndex, database);

      // Must obtain the auto-commit state outside the lock to avoid deadlocks
      // possible with some connection pool implementations
      final boolean autoCommit = isAutoCommit(connection);

      synchronized (ourLock) {
        ourConnection2RowIndex.put(connection, rowIndex);
        startNewTransaction(connection, autoCommit);
      }
    }
  }

  @MethodPattern("javax.naming.InitialContext:lookup(String)")
  public static final class InitialContext_lookup_Probe {
    public static void onReturn(
      @ReturnValue final Object returnValue,
      @Param(1) final String context
    ) {
      if (context != null && (returnValue instanceof DataSource)) {
        synchronized(ourLock) {
          ourDataSource2Context.put((DataSource)returnValue, context);
        }
      }
    }
  }

  @MethodPattern("org.apache.commons.dbcp.BasicDataSource:setUrl(String)")
  public static final class BasicDataSource_setUrl_Probe {
    public static void onReturn(
      @This final DataSource dataSource,
      @Param(1) final String url
    ) {
      if (url == null) {
        return;
      }
      synchronized(ourLock) {
        ourDataSource2Context.put(dataSource, url);
      }
    }
  }

  @MethodPattern("*:getConnection() java.sql.Connection")
  public static final class DataSource_getConnection_Probe {
    public static void onReturn(
      @This final DataSource dataSource,
      @ReturnValue final Connection connection
    ) {
      if (connection == null) {
        return;
      }

      // Must obtain the auto-commit state outside the lock to avoid deadlocks
      // possible with some connection pool implementations
      final boolean autoCommit = isAutoCommit(connection);

      synchronized (ourLock) {
        String context = ourDataSource2Context.get(dataSource);
        if (context == null) {
          // DataSource URL is unknown
          context = "DataSource@" + System.identityHashCode(dataSource);
        }
        final int rowIndex = T_CONNECTION.createRow();
        T_CONNECTION.myDatabase.setValue(rowIndex, context);
        ourConnection2RowIndex.put(connection, rowIndex);
        startNewTransaction(connection, autoCommit);
      }
    }
  }

  @MethodPattern("*:getConnection() java.sql.Connection")
  public static final class PooledConnection_getConnection_Probe {
    public static void onReturn(
      @This final PooledConnection pooledConnection,
      @ReturnValue final Connection connection
    ) {
      if (connection == null) {
        return;
      }

      if (pooledConnection instanceof DataSource) {
        return;
      }

      // Must obtain the auto-commit state outside the lock to avoid deadlocks
      // possible with some connection pool implementations
      final boolean autoCommit = isAutoCommit(connection);

      synchronized (ourLock) {
        final String context = "PooledConnection@" + System.identityHashCode(pooledConnection);
        final int rowIndex = T_CONNECTION.createRow();
        T_CONNECTION.myDatabase.setValue(rowIndex, context);
        ourConnection2RowIndex.put(connection, rowIndex);
        startNewTransaction(connection, autoCommit);
      }
    }
  }

  @MethodPattern(
    {
      "*:execute(String, int) boolean",
      "*:execute(String) boolean",
      "*:execute(String, int[]) boolean",
      "*:execute(String, String[]) boolean",
      "*:executeQuery(String) java.sql.ResultSet",
      "*:executeUpdate(String, int[]) int",
      "*:executeUpdate(String) int",
      "*:executeUpdate(String, int) int",
      "*:executeUpdate(String, String[]) int"
    }
  )
  public static final class Statement_execute_Probe {
    public static int onEnter(
      @This final Statement statement,
      @Param(1) final String query
    ) {
      if (statement instanceof PreparedStatement) {
        return -1;
      }

      final int statementRowIndex = ourStatement2RowIndex.get(statement);
      if (statementRowIndex == Table.NO_ROW) {
        return -1;
      }

      final int rowIndex = T_QUERY.createRow(statementRowIndex);
      T_QUERY.mySQL.setValue(rowIndex, query == null ? "<unknown>" : query);
      return rowIndex;
    }

    public static void onReturn(@OnEnterResult final int rowIndex) {
      if (rowIndex != -1) {
        T_QUERY.closeRow(rowIndex);
      }
    }
  }

  @MethodPattern(
    {
      "*:execute() boolean",
      "*:executeQuery() java.sql.ResultSet",
      "*:executeUpdate() int"
    }
  )
  public static final class PreparedStatement_execute_Probe {
    public static int onEnter(@This final PreparedStatement statement) {
      final int statementRowIndex = ourStatement2RowIndex.get(statement);
      if (statementRowIndex == Table.NO_ROW) {
        return -1;
      }

      return T_PREPARED_STATEMENT_QUERY.createRow(statementRowIndex);
    }

    public static void onReturn(@OnEnterResult final int rowIndex) {
      if (rowIndex != -1) {
        T_PREPARED_STATEMENT_QUERY.closeRow(rowIndex);
      }
    }
  }

  @MethodPattern("*:addBatch(String) void")
  public static final class Statement_addBatch_Probe {
    public static int onEnter(
      @This final Statement statement,
      @Param(1) final String query
    ) {
      if (statement instanceof PreparedStatement) {
        return -1;
      }

      final int statementRowIndex = ourStatement2RowIndex.get(statement);
      if (statementRowIndex == Table.NO_ROW) {
        return -1;
      }

      final int rowIndex = T_BATCH.createRow(statementRowIndex);
      T_BATCH.mySQL.setValue(rowIndex, query == null ? "<unknown>" : query);
      T_BATCH.myOperation.setValue(rowIndex, BATCH_OPERATION_ADD);
      return rowIndex;
    }

    public static void onReturn(@OnEnterResult final int rowIndex) {
      if (rowIndex != -1) {
        T_BATCH.closeRow(rowIndex);
      }
    }
  }

  @MethodPattern("*:addBatch() void")
  public static final class PreparedStatement_addBatch_Probe {
    public static int onEnter(@This final PreparedStatement statement) {
      final int statementRowIndex = ourStatement2RowIndex.get(statement);
      if (statementRowIndex == Table.NO_ROW) {
        return -1;
      }

      final int rowIndex = T_PREPARED_STATEMENT_BATCH.createRow(statementRowIndex);
      T_PREPARED_STATEMENT_BATCH.myOperation.setValue(rowIndex, BATCH_OPERATION_ADD);
      return rowIndex;
    }

    public static void onReturn(@OnEnterResult final int rowIndex) {
      if (rowIndex != -1) {
        T_PREPARED_STATEMENT_BATCH.closeRow(rowIndex);
      }
    }
  }

  @MethodPattern("*:executeBatch() int[]")
  public static final class Statement_executeBatch_Probe {
    public static int onEnter(@This final Statement statement) {
      if (statement instanceof PreparedStatement) {
        return -1;
      }

      final int statementRowIndex = ourStatement2RowIndex.get(statement);
      if (statementRowIndex == Table.NO_ROW) {
        return -1;
      }

      final int rowIndex = T_BATCH.createRow(statementRowIndex);
      T_BATCH.mySQL.setValue(rowIndex, "");
      T_BATCH.myOperation.setValue(rowIndex, BATCH_OPERATION_EXECUTE);
      return rowIndex;
    }

    public static void onReturn(@OnEnterResult final int rowIndex) {
      if (rowIndex != -1) {
        T_BATCH.closeRow(rowIndex);
      }
    }
  }

  @MethodPattern("*:executeBatch() int[]")
  public static final class PreparedStatement_executeBatch_Probe {
    public static int onEnter(@This final PreparedStatement statement) {
      final int statementRowIndex = ourStatement2RowIndex.get(statement);
      if (statementRowIndex == Table.NO_ROW) {
        return -1;
      }

      final int rowIndex = T_PREPARED_STATEMENT_BATCH.createRow(statementRowIndex);
      T_PREPARED_STATEMENT_BATCH.myOperation.setValue(rowIndex, BATCH_OPERATION_EXECUTE);
      return rowIndex;
    }

    public static void onReturn(@OnEnterResult final int rowIndex) {
      if (rowIndex != -1) {
        T_PREPARED_STATEMENT_BATCH.closeRow(rowIndex);
      }
    }
  }

  @MethodPattern("*:clearBatch() void")
  public static final class Statement_clearBatch_Probe {
    public static int onEnter(@This final Statement statement) {
      if (statement instanceof PreparedStatement) {
        return -1;
      }
      final int statementRowIndex = ourStatement2RowIndex.get(statement);
      if (statementRowIndex == Table.NO_ROW) {
        return -1;
      }

      final int rowIndex = T_BATCH.createRow(statementRowIndex);
      T_BATCH.mySQL.setValue(rowIndex, "");
      T_BATCH.myOperation.setValue(rowIndex, BATCH_OPERATION_CLEAR);
      return rowIndex;
    }

    public static void onReturn(@OnEnterResult final int rowIndex) {
      if (rowIndex != -1) {
        T_BATCH.closeRow(rowIndex);
      }
    }
  }

  @MethodPattern("*:clearBatch() void")
  public static final class PreparedStatement_clearBatch_Probe {
    public static int onEnter(@This final PreparedStatement statement) {
      final int statementRowIndex = ourStatement2RowIndex.get(statement);
      if (statementRowIndex == Table.NO_ROW) {
        return -1;
      }

      final int rowIndex = T_PREPARED_STATEMENT_BATCH.createRow(statementRowIndex);
      T_PREPARED_STATEMENT_BATCH.myOperation.setValue(rowIndex, BATCH_OPERATION_CLEAR);
      return rowIndex;
    }

    public static void onReturn(@OnEnterResult final int rowIndex) {
      if (rowIndex != -1) {
        T_PREPARED_STATEMENT_BATCH.closeRow(rowIndex);
      }
    }
  }

  @MethodPattern(
    {
      "*:close() void",
      "-*Stream*:*()",
      "-*Socket*:*()",
      "-*File*:*()"
    }
  )
  public static final class Connection_close_Probe {
    public static void onReturn(@This final Connection connection) {
      final int rowIndex;
      synchronized (ourLock) {
        rowIndex = ourConnection2RowIndex.remove(connection);
        if (rowIndex == Table.NO_ROW) {
          return;
        }

        closeExistingTransaction(connection);
      }

      T_CONNECTION.closeRow(rowIndex);
    }
  }

  @MethodPattern(
    {
      "*:close() void",
      "-*Stream*:*()",
      "-*Socket*:*()",
      "-*File*:*()"
    }
  )
  public static final class Statement_close_Probe {
    public static void onReturn(@This final Statement _this) {
      final int rowIndex = ourStatement2RowIndex.remove(_this);
      if (rowIndex == Table.NO_ROW) {
        return;
      }

      if (_this instanceof PreparedStatement) {
        T_PREPARED_STATEMENT.closeRow(rowIndex);
      }
      else {
        T_STATEMENT.closeRow(rowIndex);
      }
    }
  }

  @MethodPattern("*:commit() void")
  public static final class Connection_commit_Probe {
    public static void onReturn(@This final Connection connection) {
      // Must obtain the auto-commit state outside the lock to avoid deadlocks
      // possible with some connection pool implementations
      final boolean autoCommit = isAutoCommit(connection);

      synchronized (ourLock) {
        closeExistingTransaction(connection);
        startNewTransaction(connection, autoCommit);
      }
    }
  }

  @MethodPattern(
    {
      "*:createStatement(int, int, int) java.sql.Statement",
      "*:createStatement(int, int) java.sql.Statement",
      "*:createStatement() java.sql.Statement"
    }
  )
  public static final class Connection_createStatement_Probe {
    public static void onReturn(
      @This final Connection connection,
      @ReturnValue final Statement statement
    ) {
      if (statement == null) {
        return;
      }
      synchronized (ourLock) {
        if (ourStatement2RowIndex.containsKey(statement)) {
          return;
        }

        final int connectionRowIndex = ourConnection2RowIndex.get(connection);
        if (connectionRowIndex == Table.NO_ROW) {
          return;
        }

        final int rowIndex = T_STATEMENT.createRow(connectionRowIndex);
        ourStatement2RowIndex.put(statement, rowIndex);
      }
    }
  }

  @MethodPattern(
    {
      "*:prepareCall(String, int, int, int)",
      "*:prepareCall(String, int, int)",
      "*:prepareCall(String)",
      "*:prepareStatement(String, int)",
      "*:prepareStatement(String, String[])",
      "*:prepareStatement(String, int, int, int)",
      "*:prepareStatement(String)",
      "*:prepareStatement(String, int, int)",
      "*:prepareStatement(String, int[])"
    }
  )
  public static final class Connection_prepareStatement_Probe {
    public static void onReturn(
      @This final Connection connection,
      @Param(1) final String query,
      @ReturnValue final PreparedStatement preparedStatement
    ) {
      if (preparedStatement == null) {
        return;
      }

      synchronized (ourLock) {
        if (ourStatement2RowIndex.containsKey(preparedStatement)) {
          return;
        }

        final int connectionRowIndex = ourConnection2RowIndex.get(connection);
        if (connectionRowIndex == Table.NO_ROW) {
          return;
        }

        final int rowIndex = T_PREPARED_STATEMENT.createRow(connectionRowIndex);
        T_PREPARED_STATEMENT.mySQL.setValue(rowIndex, query);
        ourStatement2RowIndex.put(preparedStatement, rowIndex);
      }
    }
  }

  private static void closeExistingTransaction(final Connection connection) {
    final int rowIndex = ourConnection2TransactionID.remove(connection);
    if (rowIndex != Table.NO_ROW) {
      T_TRANSACTION.closeRow(rowIndex);
    }
  }

  private static void startNewTransaction(final Connection connection, final boolean isAutoCommit) {
    if (isAutoCommit) {
      return;
    }

    // Start new transaction
    final int connectionRowIndex = ourConnection2RowIndex.get(connection);
    if (connectionRowIndex == Table.NO_ROW) {
      return;
    }

    final int newRowIndex = T_TRANSACTION.createRow(connectionRowIndex);
    ourConnection2TransactionID.put(connection, newRowIndex);
  }

  private static boolean isAutoCommit(final Connection connection) {
    try {
      return connection.getAutoCommit();
    }
    catch (final Throwable ignored) {
      return true; // auto-commit default value
    }
  }
}
